/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/
#ifndef _MENU_H_
#define _MENU_H_

#include <inttypes.h>

extern uint8_t rc5_code;

typedef struct
{
    void * onPixel;
    void * onTimeChanged;
    void * onKey;
    void * onEnter;
    uint8_t use_fifo;
} menuitem_t;

typedef struct
{
    int8_t item_id;
    int8_t sel_id;
    
    uint8_t time_changed;
    
    menuitem_t item;
    void * pdata;
    union
    {
        struct
        {
            struct text_t   * menu;
            struct uhr_t    * uhr;
            struct text_t   * text;
            struct test_t   * test;
            struct snake_t  * snake;
        };
        void * addr[0];
    } data;
} menu_t;

extern menu_t menu;

#define OFFSETOF(X,Y) (((int) &X.Y-((int) &X))/sizeof(X.Y))

enum
{
    MENUID_MENU  = OFFSETOF(menu.data, menu),
    MENUID_UHR   = OFFSETOF(menu.data, uhr),
    MENUID_TEXT  = OFFSETOF(menu.data, text),
    MENUID_TEST  = OFFSETOF(menu.data, test),
    MENUID_SNAKE = OFFSETOF(menu.data, snake),
    
    MENUIDS = sizeof (menu.data) / sizeof (menu.data.menu)
};

extern void __attribute__((noreturn)) menu_main (void);

typedef void (*onEnter_t)       (void*);
typedef void (*onKey_t)         (void*,uint8_t);
typedef void (*onPixel_t)       (void*);
typedef void (*onTimeChanged_t) (void*,long);

#include "vektor-zeichen.h"

extern letter_drawinfo_t drawinfo;

#endif /* _MENU_H_ */
