/*
    This file is part of SNAKE.

    SNAKE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    SNAKE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#ifndef _SNAKE_H_
#define _SNAKE_H_

#include <inttypes.h>

enum
{
    SN_HEAD, // Kopf
    SN_FOOD  // Futter
};

#define SEG_MAX 25
    
typedef struct
{
    // Verzoegerung zwischen Aufrufen von onTimeChanged
    uint8_t delay;
    // Soviel waechst das Vieh wenns was frisst
    uint8_t grow;
    // Alles bewegt sich auf einem Raster, also mod 1, 2, 4, 8, ...
    uint8_t raster;
    // Darf neues Futter auf einem Segment platziert werden?
    uint8_t food_on_snake;
} snake_level_t;

typedef struct snake_t
{
    // Kriech-Richtung
    uint8_t dir;
    // Neue Richtung angefordert
    uint8_t new_dir;
    
    // Futter laesst die Schlong wachsen
    uint8_t grow;

    // einzelne Punkte: Kopf, Futter, ...
    union
    {
        struct
        {
            struct { uint8_t x, y; } head;
            struct { uint8_t x, y; } food;
        };
        struct { uint8_t x, y; } p[0];
    };
    
    // Wird gerade ein neuer Platz fuers Futter gesucht??
    uint8_t seek_food;
    // Hilfsvariable dabei
    uint8_t food_ok;
    
    // Fuer Umlaufzahl Mauer um Futter
    int8_t wall_umlauf;
    
    // # zu zeichnender Texte
    uint8_t ntexte;
    
    // # y-Pos fuer "Game Over"
    uint8_t scroll_y;
    // # y-Pos fuer "Game Over"
    uint8_t delay_scroll;
    
    // # zu zeichnender Texte
    uint8_t delay;
    
    struct
    {
        // momentan gezeichnetes Element
        // -2, -1: head
        // 0..nseg-1: body
        uint8_t elem;
        // Anzahl noch zu zeichnende Pixel des aktuellen Elements
        uint8_t len;
        // momentane Zeichenrichtung
        uint8_t dir;
        // Momentane Strahlposition
        uint8_t x, y;
        // Alles fertig gemalt
        uint8_t done;
        // Kollisionen
        const uint8_t * wall;
        
        uint8_t bums;
        
        // Nummer des gerade gezeichneten Textes
        uint8_t ntext;
    } draw;
    
    struct
    {
        uint8_t snake;
        uint8_t text;
    } changed;
    
    uint8_t raster_mask;
    
    struct
    {
        // Eingestellter Level: 0..
        uint8_t level;
        // Erreichte Punktzahl
        uint16_t score;
        // Spiel ist aus :-(
        uint8_t over;
    } game;
    
    snake_level_t level;
    
    // Koerper
    struct
    {
        // Index des ersten Segments
        uint8_t start;
        // Anzahl Segmente
        uint8_t nseg;
        // Segmente
        struct
        {
            // Richtung + Laenge
            uint8_t dir, len;
        } seg[SEG_MAX];
    } body;
} snake_t;

extern snake_t snake;
extern void snake_set_level (snake_t*, uint8_t);

#include "menu.h"
extern const menuitem_t menuitem_snake;

#endif /* _SNAKE_H_ */
