/*
    This file is part of VEKFONT.

    VEKFONT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    VEKFONT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include "vektor-zeichen.h"

#if defined (__GNUC__) && defined (__AVR__)
    // Sonderlocken fr avg-gcc
    #include "pgmread-inc.h"
#else
    // Fr normale Compiler
    #define PROGMEM
    #define pgm_read_byte_inc(addr) (*addr++)
#endif // avr-gcc

// Masken fr einige Flags
#define RIGHT1  (1 << 5)
#define LEFT1   (1 << 6)
#define SHRINKY (1 << 7)

// Falls draw_letters() nicht bis zum Ende aufgerufen wird,
// mssen die internen Daten von Hand rckgesetzt werden!
void draw_reset (letter_drawinfo_t * di)
{
    di->iPixel = 0;
    di->iLine  = 0;
    di->itext  = 0;
    di->inject_char = 0;
}

// Gibt den Text in di->text ab der Startposition di->start aus,
// indem die Callback-Funktionen di->beam(x,y) bzw di->beam->skip()
// aufgerufen werden.
//
// !!! Die Funktionen mssen initialisiert sein, ansonsten landet das
// !!! Programm im Nirvana.
//
// return-Wert
// 1: Es wurde kein Pixel ausgegeben, die Textausgabe ist fertig
// 0: Es wurde genau 1 Pixel via di->beam_xy(x,y) ausgegeben.
//    Evtl. wurde vor der Pixelausgabe via di->beam_skip()
//    ber einen Sprung der Ausgabe-Position informiert.
//    Falls di->beam_skip() nicht aufgerufen wurde, dann ist
//    der Abstand zum vorhergehenden Pixel (Maximums-Norm)
//    hchstens Eins, in der Euklidischen Norm maximal \sqrt{2}.

uint8_t draw_letters (letter_drawinfo_t * di)
{
    while (0==0)
    {
        uint8_t iLine = di->iLine;

        if (di->iPixel == 0)
        {
            // Neue Linie fngt an
            if (iLine == 0)
            {
                // Neues Zeichen fngt an
                uint8_t zeichen = di->inject_char;

                if (!zeichen)
                {
                    vf_uint_t itext = di->itext;
                    if (itext == 0)
                    {
                        // Erstes Zeichen: Startposition
                        di->p.x = di->start.x;
                        di->p.y = di->start.y;
                    }

                    const char * pzeichen = &di->text[itext];

#if defined (__GNUC__) && defined (__AVR__)
                    if (di->text_in_flash)
                        zeichen = pgm_read_byte (pzeichen);
                    else
#endif // avr-gcc
                        zeichen = (uint8_t) *pzeichen;

                    di->itext = itext+1;
                }

                if (zeichen == '\0')
                {
                    // Stringende: fertig
                    di->itext = 0;
                    return 1;
                }

#ifdef VF_NOTIFY_CHAR
                {
                    // Informiert ber Anfang eines neuen Zeichens,
                    // wird nur fr SVG-Ausgabe benutzt.
                    extern void VF_NOTIFY_CHAR (letter_drawinfo_t *, char);
                    VF_NOTIFY_CHAR (di, (char) zeichen);
                }
#endif // NOTIFY_CHAR

                uint8_t zoom = di->zoom;

                if (zeichen == '\r')
                    continue;
                    
                // Zeilenumbruch
                if (zeichen == '\n')
                {
                    di->p.y -= (vf_uint_t) VF_SPACE_Y(zoom);
                    di->p.x = di->start.x;
                    continue;
                }

                // Leerzeichen
                if (zeichen == ' ')
                {
                    di->p.x += VF_SPACE_X(zoom);
                    continue;
                }

#if (VEKFONT & VF_MAP_UMLAUTS) && !(VEKFONT & VF_UMLAUTS)
                // Umlaute umwandeln
                uint8_t inject = 0;
                char c = zeichen;
                if (c == '')    zeichen = 'A', inject = 'e';
                if (c == '')    zeichen = 'O', inject = 'e';
                if (c == '')    zeichen = 'U', inject = 'e';
                if (c == '')    zeichen = 'a', inject = 'e';
                if (c == '')    zeichen = 'o', inject = 'e';
                if (c == '')    zeichen = 'u', inject = 'e';
                if (c == '')    zeichen = 's', inject = 's';
                di->inject_char = inject;
#endif // MAP_UMLAUTS

#if (VEKFONT & VF_MAP_LOWERCASE) && !(VEKFONT & VF_LOWERCASE)
                // Kleinbuchstaben umwandeln
                if (zeichen >= 'a' && zeichen <= 'z')
                    zeichen += 'A'-'a';
#endif // MAP_LOWERCASE

                di->pix.y = di->p.y;
                di->pix.x = di->p.x;

                // Zeichen suchen
                const uint8_t * addr = vektor_zeichen_data;

#if VF_LOOKUP_END > VF_LOOKUP_START
                // In lookup-Tabelle nachschauen, falls vorhanden
                static const uint8_t * lookup[(uint8_t) (VF_LOOKUP_END - VF_LOOKUP_START +1)];

                const uint8_t ** lookup_addr = 0;
                
                if (zeichen >= VF_LOOKUP_START 
                    && zeichen <= VF_LOOKUP_START + VF_LOOKUP_END)
                {
                    lookup_addr = & lookup[zeichen - VF_LOOKUP_START];
                
                    if (*lookup_addr)
                        addr = *lookup_addr;
                }
#endif // LOOKUP

                uint8_t nLines;
                uint8_t nLinesRightLeft;
                uint8_t id;

                // Zeichen im Zeichensatz suchen. Falls das Zeichen in der
                // lookup-Tabelle eingetragen ist, wird es direkt gefunden
                while (1)
                {
                    id = pgm_read_byte_inc (addr);
                    nLinesRightLeft = pgm_read_byte_inc (addr);

                    nLines = nLinesRightLeft & 0x1f;
                    
                    if (id == zeichen)
                        break;
                        
                    if (id == VF_UNKNOWN)
#ifdef VF_EXTRA_GLYPHS
                    {
                        extern const uint8_t VF_EXTRA_GLYPHS[];
                        addr = (const uint8_t*) VF_EXTRA_GLYPHS;
                        continue;
                    }
                    if (id == VF_UNKNOWN_EXTRA)
#endif // VF_EXTRA_GLYPHS
                        break;
                        
                    // berspringe diesen Buchstaben
                    addr += nLines;
                }

                // Zeichen gefunden
                di->zeichen = id;
                di->addr    = addr;
                di->nLines  = nLines;

#if VF_LOOKUP_END > VF_LOOKUP_START
                if (lookup_addr)
                    *lookup_addr = addr - 2;
#endif // LOOKUP

                uint8_t proportional = 0;

#if VEKFONT & VF_PROPORTIONAL
                proportional = di->proportional;
#endif // PROPORTIONAL

#if VEKFONT & VF_LOWERCASE
                if (proportional)
                {
                    // T und F unterschneiden
                    if (id == 'T' || id == 'F')
                        nLinesRightLeft += LEFT1;
                }
#endif // LOWERCASE

                di->left  = nLinesRightLeft;
                di->right = 0;

                if ((nLinesRightLeft & RIGHT1) && !proportional)
                    di->right = zoom;

#if VEKFONT & VF_SHRINKY
                di->shrinky = 0;
                di->shrink.x = 0;
                di->shrink.y = 0;
#endif // SHRINKY

            } // iLine == 0

            const uint8_t * addr = di->addr;

            // Lese Daten, Lesezeiger eins weiter
            uint8_t data = pgm_read_byte_inc (addr);

            // Dekodiere LINE-Daten
            vf_uint_t nPixels = data >> 5;

            vf_int_t dx, dy;
            uint8_t zoom = di->zoom;

            if (nPixels == 0)
            {
                // GOTO: Setzt neue x-y-Position
                dx = data;
                dy = pgm_read_byte_inc (addr);

                di->addr = addr;

#if VEKFONT & VF_SHRINKY
                if (zoom > 1)
                    di->shrinky = dy & SHRINKY;
                dy &= ~SHRINKY;
#endif // SHRINKY

                dx = (vf_uint_t) dx * zoom;
                dy = (vf_uint_t) dy * zoom;

                dx += di->right;

                di->pix.x = di->p.x + dx;
                di->pix.y = di->p.y + dy;

                di->beam_skip();
                di->iLine += 2;

                continue;
            }

            di->addr = addr;

            if (di->iPixel == 0)
            {
                if (iLine == 0)
                {
                    // Neues Zeichen, das an 0,0 beginnt --> skip
                    di->pix.x += di->right;
                    di->beam_skip ();
                }
            }

            //  LINE

            nPixels *= zoom;
#if VF_PLOTTER == 2
            if (data != DOTS_L && data != DOTS_R)
            {
                nPixels++;
                di->noskip = 1;
            }
            else
#endif // PLOTTER
            {
                nPixels--;

                uint8_t d;

                // Ende zeichnen
                if ((d = data & 1, d))  nPixels++;

                // Start zeichnen
                if ((d = data & 2, d))  nPixels++, di->noskip = 1;

                if (!nPixels)
                {
                    di->noskip = 0;
                    goto next_pixel;
                }
            }

#if VEKFONT & (VF_UMLAUTS | VF_PUNKT | VF_DOTS)
            uint8_t dots = 0;

            if (data == DOTS_L)         di->pix.x --, dots = 1;
            if (data == DOTS_R)         di->pix.x ++, dots = 1;
            if (data == DOTS_B)         dy = -1,      dots = 2;
            if (data == DOTS_T)         dy =  1,      dots = 3;

            if (dots == 1)
            {
                di->nPixels = zoom;
                dx = 0;
                dy = 1;
                di->noskip = 0;
            }
            else
            if (dots >= 2)
            {
                if (dots == 2 && zoom == 1)
                    goto next_pixel;

                dx = 0;
                di->nPixels = (uint8_t) (zoom+1)/2;
                di->noskip = 0;
            }
            else
#endif // UMLAUTS | PUNKT
            {
                di->nPixels = nPixels;

                // dx, dy aus dxy extrahieren
                uint8_t dxy = (data >> 2) & 3;

                dy = dxy;
                if (dy)            dy = -1;
                if (dxy == 3)      dxy = 4;
                dx = dxy >> 1;
                dx -= 1;

                if (data & (1 << 4))
                    dx = -dx, dy = -dy;
            }

            di->d.x = dx;
            di->d.y = dy;
        } // di->iPixel == 0

        vf_uint_t pixx = di->pix.x;
        vf_uint_t pixy = di->pix.y;

        if (!di->noskip)
        {
            pixx += di->d.x;
            pixy += di->d.y;
        }
        di->noskip = 0;

        // Pixel zeichnen
        di->pix.x = pixx;
        di->pix.y = pixy;

        uint8_t doplot = 1;

#if VEKFONT & VF_SHRINKY
        if (di->shrinky)
        {
            uint8_t y = pixy - di->p.y;
            y *= 3; y /= 4;
            pixy = di->p.y + y;

            if (di->shrink.y != pixy || di->shrink.x != pixx)
            {
                di->shrink.y = pixy;
                di->shrink.x = pixx;
            }
            else
                doplot = 0;
        }
#endif // SHRINKY

        if (doplot)
#if VF_PLOTTER
            if (di->iPixel == 0 || di->iPixel == di->nPixels-1)
#endif // PLOTTER
                di->beam_xy (pixx, pixy);

        // nchstes Pixel
        if (++di->iPixel == di->nPixels)
        {
            uint8_t go_redraw = 0;

            if (0)
            {
                next_pixel:
                    go_redraw = 1;
            }

            di->iPixel = 0;

            // Linien-Index eins weiter
            if (++iLine == di->nLines)
            {
                iLine = 0;

                vf_uint_t px = di->p.x;
                uint8_t zoom = di->zoom;
                px += VF_SPACE_X(zoom);

#if VEKFONT & VF_PROPORTIONAL
                if (di->proportional)
                {
                    uint8_t left = di->left & 0xc0;
                    if (left)
                    {
                        left >>= 6;
                        left = (vf_uint_t) (left * zoom);
                        if (di->zeichen == 'c')
                            left++;
                        px -= left;
                        di->left = 0;
                    }
                }
#endif // PROPORTIONAL

                di->right = 0;
                di->p.x = px;
            }
            di->iLine = iLine;

            if (go_redraw)
                continue;
        } // ++di->iPixel == di->nPixels

        break;
    }

    return 0;
}
